// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CCollection implements common functionality of all collection object
// 		classes. The collection object classes will use this class as a member 
// 		variable instead of inheriting from it to avoid double inheritance dilemmas.
// 		Is meant to work like SAFEARRAYS

#include <stdlib.h>
#include "precomp.h"
#include "enumvar.h"

// Constructor for CNode
// CNode is allows the CCollect to maniulate different classes in a 
// generic manner.

CNode::CNode()
{
	m_pDispatch = NULL;
	m_bstrName = NULL;
	m_dwID = 0;
	m_pNext = NULL;
	m_pPrev = NULL;
	m_pNextSymbol = NULL;
	m_pPrevSymbol = NULL;
	m_pNextID = NULL;
	m_pPrevID = NULL;
}

// Destructor for CNode

CNode::~CNode()
{
	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);
		
	if(SUCCEEDED(CheckDispatch(m_pDispatch)))
		m_pDispatch->Release();
}            

// Constructor for CCollection
// Implements generic Automation collection

CCollection::CCollection()
{
   	m_cElements = 0;
	m_pFirstNode = NULL;
	m_pLastNode = NULL;
	m_pEnumVariant = NULL;

	// Create Has
	CHashTable::Create(TRUE, &m_phashSymbols);
	CHashTable::Create(FALSE, &m_phashIDs);
}

CCollection::~CCollection()
{
 	RemoveAll();
                         
	if(SUCCEEDED(CheckPointer(m_phashSymbols)))
		delete m_phashSymbols;

	if(SUCCEEDED(CheckPointer(m_phashIDs)))
		delete m_phashIDs;

	if(SUCCEEDED(CheckPointer(m_pEnumVariant)))
		m_pEnumVariant->Release();
}


/*
 * CCollection::AddObject
 *
 * Purpose:
 *	Adds a new object to the collection. 
 *
 * Parameters:
 *  pDispatch			Object to add.
 *	bstrName			Name of item to add
 *  dwID				ID of item to add
 *  ppNode				(Optional) Pointer to return created CNode
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::AddObject(LPDISPATCH pDispatch, BSTR bstrName, BOOL bRegID
		, DWORD dwID, CNode FAR* FAR* ppNode){
	
	CNode FAR *pNode = NULL;
	TCHAR	buf[MAX_DIGITS]; // for converting numbers to strings
		

	TRY{                     
		if(ppNode != NULL){
			if(FAILED(CheckPointer((LPVOID) ppNode, 0))){
    			AddInDebugOutput(NULL, TEXT("CCollection::AddObject"), IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
		
			*ppNode = NULL;
		}
        
        // If Name is invalid then return                        
        if(bstrName == NULL){
    		AddInDebugOutput(NULL, TEXT("CCollection::AddObject"), IDS_DEBUG_INVALIDNAME, NULL); 
			MYRAISEEXCEPTION;
        }                        

		// Allocate a new Node
		pNode = new CNode();
		
		if(pNode == NULL){
    		AddInDebugOutput(NULL, TEXT("CCollection::AddObject"), IDS_DEBUG_NODENOTALLOCED, NULL); 
			MYRAISEEXCEPTION;
		}

		// Fill in node with object's pDispatch
		// and searching criteria
		pNode->m_pDispatch = pDispatch;
	    pNode->m_bstrName = SysAllocString(bstrName);
	    pNode->m_dwID = dwID;
	    
	    // Insert Name into internal hash table for quick look-ups
	    if(FAILED(m_phashSymbols->Insert(bstrName, pNode))){
    		AddInDebugOutput(NULL, TEXT("CCollection::AddObject")
    			, IDS_DEBUG_HASHTABLEINSERTFAIL, NULL); 
			MYRAISEEXCEPTION;
	    }


		if(bRegID){
			memset(buf, 0, sizeof(buf));

			#ifdef UNICODE
				_ultow(pNode->m_dwID, buf, 10);
			#else
				_ultoa(pNode->m_dwID, buf, 10);
			#endif

    		if(FAILED(m_phashIDs->Insert(buf, pNode))){
				AddInDebugOutput(NULL, TEXT("CCollection::AddObject")
					, IDS_DEBUG_HASHTABLEINSERTFAIL, NULL); 
				MYRAISEEXCEPTION;
			}
	    }

	    // Insert Name into intenal link-list
	    // there is really no need to do this if you use hash tables
		if(m_cElements == 0){
			m_pFirstNode = m_pLastNode = pNode;
			pNode->m_pPrev = NULL;
		}else{
			m_pLastNode->m_pNext = pNode;
			pNode->m_pPrev = m_pLastNode;	
			m_pLastNode = m_pLastNode->m_pNext;
		}
                                                    
		m_pLastNode->m_pNext = NULL;
        
        // Collection is a user so AddRef
		pDispatch->AddRef();

		// Increment elements count
		m_cElements++;
        
        // if caller asked for pNode return it
		if(ppNode)
			*ppNode = pNode;

		return S_OK;
	}

	EXCEPT{                        
		// Clean-up
		if(pNode)
			delete pNode;

   		AddInDebugOutput(NULL, TEXT("CCollection::AddObject")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}



/*
 * CCollection::RemoveObject
 *
 * Purpose:
 *	Removes object from the collection. 
 *
 * Parameters:
 *  pDispatch			Object to remove.
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::RemoveObject(LPDISPATCH pDispatch){
	CNode *pNode = NULL;
	ULONG i;

	TRY{ 
		// Find pNode - could use separate hash table 
		for(pNode = m_pFirstNode, i=0; i<m_cElements; pNode = pNode->m_pNext, i++){
			if(pNode->m_pDispatch == pDispatch){
				break;
			}					
		}

		if(pNode == NULL){
			AddInDebugOutput(NULL, TEXT("CCollection::RemoveObject")
				, IDS_DEBUG_FINDOBJECTINCOLLECTIONFAILED, NULL); 
			MYRAISEEXCEPTION;
		}                             

		// Remove from linked list
		if(i==0){
			m_pFirstNode = m_pFirstNode->m_pNext;
		}else if(i == m_cElements - 1){
			m_pLastNode = m_pLastNode->m_pPrev;
			m_pLastNode->m_pNext = NULL;
		}else{
	 		pNode->m_pPrev->m_pNext = pNode->m_pNext;	
		}
	    
		// Remove item from hash
	    if(FAILED(m_phashSymbols->Remove(pNode->m_bstrName))){
    		AddInDebugOutput(NULL, TEXT("CCollection::RemoveObject"), IDS_DEBUG_NODENOTALLOCED, NULL); 
			MYRAISEEXCEPTION;
	    }

		// delete pNode
		delete(pNode);
		
        // decrement elements count
		m_cElements--;

		return S_OK;
	}

	EXCEPT{
		AddInDebugOutput(NULL, TEXT("CCollection::RemoveObject")
				, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}

/*
 * CCollection::FindObjectByIndex
 *
 * Purpose:
 *	Finds an object in the collection by its index. 
 *
 * Parameters:
 *  ulIndex			Object to find.                                
 *	ppDispatch		Pointer to return IDispatch of found object
 *
 * Return Value:
 *  HRESULT
 *
 * Comments:
 *	This function could be sped up by adding an offset scheme into the list
 *
 *	Requires its own function because numerical indexes sent to FindObject 
 *  Represent its ID.
 */ 

HRESULT CCollection::FindObjectByIndex(ULONG ulIndex, LPDISPATCH FAR* ppDispatch){
    CNode FAR* pNode = NULL; 
	ULONG i;

	TRY{
		if(ppDispatch != NULL){
			if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    			AddInDebugOutput(NULL, TEXT("CCollection::FindObjectByIndex"), IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
			*ppDispatch = NULL;
		}
		

		// If collection is empty immediately return
		if(m_cElements == 0){
			MYRAISEEXCEPTION;
		}
	    
	    // Make sure index is valid, else return
		if(ulIndex >= m_cElements || ulIndex < 0){
			AddInDebugOutput(NULL, TEXT("CCollection::FindObjectByIndex")
					, IDS_DEBUG_INVALIDINDEX, NULL); 
			MYRAISEEXCEPTION;
		}

		//  Move to correvt node in list; Again offset sscheme is faster
		for(pNode = m_pFirstNode, i=0; i<ulIndex; pNode = pNode->m_pNext, i++)
		{}
        
        // if ppDispatch is valid AddRef and return object's IDispatch
		if(ppDispatch){
			pNode->m_pDispatch->AddRef();
			*ppDispatch = pNode->m_pDispatch;
		}

		return S_OK;
	}
			
	EXCEPT{ 
		return ResultFromScode(E_FAIL);
	}
}



/*
 * CCollection::FindObject
 *
 * Purpose:
 *	Finds an object in the collection by Name. 
 *
 * Parameters:
 *  bstrName		Name of object to find.                                
 *	ppDispatch		Pointer to return IDispatch of found object
 *  pulIndex		Pointer to return ordinal index of found object
 *  ppNode 			Pointer to return Node of found object
 * 
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::FindObject(BSTR bstrName, LPDISPATCH FAR *ppDispatch, ULONG FAR *pulIndex, CNode FAR* FAR* ppNode){
    CNode FAR* pNode = NULL; 

	TRY{
		if(ppDispatch != NULL){
			if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    			AddInDebugOutput(NULL, TEXT("CCollection::FindObject")
    					, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
		
			*ppDispatch = NULL;
		}
			
		if(ppNode != NULL){
			if(FAILED(CheckPointer((LPVOID) ppNode, 0))){
				*ppNode = NULL;
    			AddInDebugOutput(NULL, TEXT("CCollection::FindObject")
    					, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}

			*ppNode = NULL;
		}
	
		// If collection is empty return 
		if(m_cElements == 0){
			MYRAISEEXCEPTION;
		}									  
        
        // If name is empty return
        if(bstrName == NULL){
			AddInDebugOutput(NULL, TEXT("CCollection::FindObject")
					, IDS_DEBUG_INVALIDNAME, NULL); 
        	MYRAISEEXCEPTION;
        }                    
        
        // Find object in hash table
	    if(FAILED(m_phashSymbols->Find(bstrName, (CNode FAR* FAR*) &pNode))){
			MYRAISEEXCEPTION;
	    }
        
        // if valid dispatch is available AddRef and return object
		if(ppDispatch){
			*ppDispatch = pNode->m_pDispatch;	    
			(*ppDispatch)->AddRef();
		}
		
		// If caller asks for pNode then return it
		if(ppNode){
			*ppNode = pNode;
		}                                       
		
		return S_OK;									
	}
			
	EXCEPT{
		return ResultFromScode(E_FAIL);
	}
}

/*
 * CCollection::FindObject
 *
 * Purpose:
 *	Finds an object in the collection by ID. 
 *
 * Parameters:
 *  dwID			ID of object to find.                                
 *	ppDispatch		Pointer to return IDispatch of found object
 *  pulIndex		Pointer to return ordinal index of found object
 *  ppNode 			Pointer to return Node of found object
 * 
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::FindObject(DWORD dwID, LPDISPATCH FAR *ppDispatch
			, ULONG FAR *pulIndex, CNode FAR* FAR* ppNode){
    
    CNode FAR* pNode = NULL; 

	TRY{
		if(ppDispatch != NULL){
			if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    			AddInDebugOutput(NULL, TEXT("CCollection::FindObject")
    				, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
		
			*ppDispatch = NULL;
		}
			
		if(ppNode != NULL){
			if(FAILED(CheckPointer((LPVOID) ppNode, 0))){
    			AddInDebugOutput(NULL, TEXT("CCollection::FindObjectByIndex")
    				, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
	
			*ppNode = NULL;
		}
		
	
		// if collection is empty return
		if(m_cElements == 0){
			MYRAISEEXCEPTION;
		}
        
		// Buf is for converting numbers to strings.
		TCHAR buf[MAX_DIGITS]; // for converting numbers to strings
		
		// Make ID a string for hash look-up
		#ifdef UNICODE
			_ultow(dwID, buf, 10);
		#else
			_ultoa(dwID, buf, 10);
		#endif
	    
	    // Find object in hash
	    if(FAILED(m_phashIDs->Find(buf, &pNode))){
			MYRAISEEXCEPTION;
	    }
        
        // If ppDispatch is valid, AddRef and return object's IDispatch
		if(ppDispatch){
			*ppDispatch = pNode->m_pDispatch;	    
			(*ppDispatch)->AddRef();
		}
		
		// If caller asks for node, return it
		if(ppNode){
			*ppNode = pNode;
		}

		return S_OK;									
	}
			
	EXCEPT{
		return ResultFromScode(E_FAIL);
	}
}

/*
 * CCollection::RemoveAll
 *
 * Purpose:
 *	Clears collection. 
 *
 * Parameters:
 *	VOID
 * 
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::RemoveAll(void){
    CNode FAR* pNode = NULL; 
	ULONG i;

	TRY{
		// Remove all Nodes
		for(pNode = m_pFirstNode, i=0; i<m_cElements; pNode = pNode->m_pNext, i++){
			delete pNode;
		}
		
		// Reset count to 0
		m_cElements = 0;
	
		return S_OK;
	}

#ifdef _WIN32			
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CCollection::RemoveAll")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);
	}
#endif
}

/*
 * CCollection::__NewEnum
 *
 * Purpose:
 *	Implements common IEnumVariant
 *
 * Parameters:
 *	IUnknown *
 * 
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCollection::GetEnum(LPUNKNOWN *ppUnk){
	TRY{                                     
		if(FAILED(CheckPointer((LPVOID) ppUnk, 0))){
			return ResultFromScode(E_FAIL);
		}
	
    	if(m_pEnumVariant == NULL){
    		if (FAILED(CEnumVariant::Create(this, &m_pEnumVariant))){
				AddInDebugOutput(NULL
					, TEXT("CCollection::GetEnum")
					, IDS_DEBUG_IENUMVARIANTCREATEFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}
		}

		m_pEnumVariant->QueryInterface(IID_IUnknown, (LPVOID *) ppUnk);
		
		return S_OK;
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CCollection::RemoveAll")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);
	}
}


