// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		Implemntatioon of the add-in's DataObjects collections. This is not
// 		a pass-through object to a bound WOSA/XRT DataObjects collection. This
// 		collection is unique to the add-in and maintains the DataObjects
// 		the add-in is bound to.

#include "precomp.h"
#include "guid.h"

#include "realtime.h"
#include "dataobjs.h"
#include "dataobj.h"

#include "enumvar.h"
#include "cache.h"

#include "framewrk.h"

/*
 * CDataObjects::Create
 *
 * Purpose:
 *  Creates an instance of a DataObjects collection object 
 *  and initializes it.
 *
 * Parameters
 *	ppDataObjects	Pointer to return create DataObjects
 *
 * Return Value
 *	HRESULT
 *
 */

HRESULT CDataObjects::Create(CDataObjects FAR* FAR* ppDataObjects) 
{   
    CDataObjects FAR* pDataObjects = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppDataObjects, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObjects::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*ppDataObjects = NULL;
    
    	// Create a new collection
    	pDataObjects = new CDataObjects();

    	if (pDataObjects == NULL){
			AddInDebugOutput(&pDataObjects->m_excepinfo
					, TEXT("CDataObjects::Create")
					, IDS_DEBUG_DATAOBJECTSALLOCFAILED
					, DISP_E_EXCEPTION); 
			return ResultFromScode(E_FAIL);
    	}

   
    	// Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
    	if (FAILED(LoadTypeInfo(&pDataObjects->m_ptinfo, IID_IDataObjects))){
			AddInDebugOutput(&pDataObjects->m_excepinfo
					, TEXT("CDataObjects::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}
		
		// Get hInstance for window creation
		XLOPER xRes;
		Excel(xlGetInst, (LPXLOPER) &xRes, 0);
		HINSTANCE hInstance = (HINSTANCE) xRes.val.w;
    	        
		Excel(xlGetHwnd, (LPXLOPER) &xRes, 0);
		HWND hwndXL = (HWND) xRes.val.w;
    	        
		#ifndef _WIN32
			SetMessageQueue(112);
    	#endif

    	// Position the Window, now		
		int sx = GetSystemMetrics(SM_CXSCREEN);
		int sy = GetSystemMetrics(SM_CYSCREEN);
		int	nWidth = 400 * g_nBorderX;
		int nHeight = 50 * g_nBorderY;
		
		pDataObjects->m_hWnd = CreateWindow(TEXT("XRTXLL_CLB_RC")
			, TEXT(""), WS_POPUP | WS_DLGFRAME
			, (sx-nWidth)/2, (sy-nHeight)/2
			, nWidth, nHeight
			, hwndXL, NULL, hInstance, (LPVOID) 0L);           

  		if (pDataObjects->m_hWnd == NULL){
			AddInDebugOutput(&pDataObjects->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CREATEWINDOWFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
		}                  
				
    	*ppDataObjects = pDataObjects;

	    return S_OK;
	}
							  	
	EXCEPT{    
   		AddInDebugOutput(NULL, TEXT("CDataObjects::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 
   			
    	if (pDataObjects)
	   		delete pDataObjects;           

  		return ResultFromScode(E_FAIL);
	}
}

/*
 * CDataObjects::CDataObjects
 *
 * Purpose:
 *  Constructor for CDataObjects object. Initializes members to NULL.
 *
 */

CDataObjects::CDataObjects()
{    
	m_pRealTimeExt = NULL;                               
	m_ptinfo = NULL;
	m_cRef = 1;
   	m_bRaiseException = 0;
	m_bstrStatusWndMsg = NULL;
	m_hWnd = 0;
	m_fComplete = (float) 0.0;		
					 	
}


/*
 * CDataObjects::~CDataObjects
 *
 * Purpose:
 *  Destructor for CDataObjects object. 
 *
 */

CDataObjects::~CDataObjects()
{    
	if(m_hWnd && IsWindow(m_hWnd))
		DestroyWindow(m_hWnd);

	if(SUCCEEDED(CheckPointer(m_ptinfo)))
		m_ptinfo->Release();
}


/*
 * CDataObjects::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CDataObjects::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IDataObjects) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DDataObjects)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CDataObjects::AddRef(void)
{ 
	return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CDataObjects::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CDataObjects::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */

STDMETHODIMP
CDataObjects::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CDataObjects::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */

STDMETHODIMP
CDataObjects::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CDataObjects::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */

STDMETHODIMP 
CDataObjects::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
	  return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}																		 
													
/*
 * CDataObjects::Invoke							
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *
 */

STDMETHODIMP
CDataObjects::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

    m_bRaiseException = FALSE;
	 
    hr = DispInvoke(          
		this, m_ptinfo,
		dispidMember, wFlags, pdispparams,
		pvarResult, pexcepinfo, puArgErr);
    
    // Check if property accessor function or method called by DispInvoke wants to raise
    // an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
    // structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
    // property or method.    
    if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);
		
		m_bRaiseException = FALSE;
        
        m_excepinfo.scode = NULL;
        m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
    }else{
		pexcepinfo->scode = S_OK; 
    	return hr;    
	}
}



/*
 * CDataObjects::Add
 *
 * Purpose:
 *  Creates and adds a new DataObject to the collection
 *
 * Parameters
 *	ppDataObjects	Pointer to return create DataObjects
 *
 * Return Value
 *	HRESULT
 *
 */

STDMETHODIMP_(LPDISPATCH)   
CDataObjects::Add(LPDISPATCH pXRTDispatch, VARIANT vName, VARIANT vRange)
{ 
    CDataObject FAR* pDataObject = NULL;  
    LPDISPATCH pMyDispatch = NULL;

	TRY{
		// Check DataObject name. It must be ok.    
    	if(FAILED(CheckDispatch(pXRTDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_BADPOINTER
					, DISP_E_EXCEPTION); 
       		MYRAISEEXCEPTION;  
    	}

		if(FAILED(CheckVariant(&vName, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_INVALIDNAME
					, DISP_E_EXCEPTION); 
       		MYRAISEEXCEPTION;  
		}
    
		//Create a new DataObject
    	if (FAILED(CDataObject::Create(&pDataObject, pXRTDispatch
		   		, vName.bstrVal, vRange))){

			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Add")
					, IDS_DEBUG_DATAOBJCREATEFAILED
					, DISP_E_EXCEPTION); 
       		MYRAISEEXCEPTION;  
    	}       
    
    	//Finished with this 
    	VariantClear(&vRange);
                                          
    	///Give DataObject a pointer to top
    	pDataObject->m_pRealTimeExt = m_pRealTimeExt;		
        
		//Internal QI is just a recast
    	if(FAILED(pDataObject->QueryInterface(IID_IDispatch
    			, (LPVOID FAR *) &pMyDispatch))){

			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Add")
					, IDS_DEBUG_NODATAOBJDISPATCH
					, DISP_E_EXCEPTION); 
       		MYRAISEEXCEPTION;  
		}

        pDataObject->Release(); // Release from creation
        pDataObject = NULL;
                                       
    	// Add new DataObject to collection
		if(FAILED(m_col.AddObject(pMyDispatch, vName.bstrVal, FALSE, 0))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Add")
					, IDS_DEBUG_REQUESTSADDFAILED
					, DISP_E_EXCEPTION); 
       		MYRAISEEXCEPTION;  
		}
        
	    return pMyDispatch;
	}

	EXCEPT{
		if(pDataObject){
			delete pDataObject;
		}
    
    	VariantClear(&vRange);
		
	    m_bRaiseException = TRUE;
	    
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::Add")
   			, IDS_DEBUG_EXCEPTION, NULL); 

    	return 0L;   
	}
}


/*
 * CDataObjects::Remove
 *
 * Purpose:
 *  Removes a specific DataObject from the collection
 *
 * Parameters
 *	vParam	DataObject to remove
 *
 * Return Value
 *	HRESULT
 *
 */

STDMETHODIMP_(void)
CDataObjects::Remove(VARIANT vParam)
{   
	ULONG lIndex;
	LPDISPATCH pDispatch = NULL;
	HRESULT hr;

	TRY{
		if(FAILED(CheckVariant(&vParam, TRUE))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::Remove")
   				, IDS_DEBUG_BADVARIANT, NULL); 
			MYRAISEEXCEPTION;
		}

		// Find object based on VARIANT index
   		if(vParam.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) vParam.iVal, &pDispatch);
			lIndex = (long) vParam.iVal;
   		}else if(vParam.vt == VT_I4){
   			hr = m_col.FindObjectByIndex(vParam.lVal, &pDispatch);
			lIndex = (long) vParam.lVal;
   		}else if(vParam.vt == VT_BSTR){
   			hr = m_col.FindObject(vParam.bstrVal, &pDispatch, &lIndex);
   		}else{
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Remove")
					, IDS_DEBUG_INVALIDINDEX
					, DISP_E_EXCEPTION); 
  			MYRAISEEXCEPTION;
   		}

		// Check index                                  
		if(FAILED(hr)){        
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Remove")
					, IDS_DEBUG_DATAOBJNOTFOUNDINCOLLECTION
					, DISP_E_EXCEPTION); 
  			MYRAISEEXCEPTION;
		}

		// Update collection
		if(FAILED(m_col.RemoveObject(pDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::Remove")
					, IDS_DEBUG_COLLECTIONREMOVEFAILED
					, DISP_E_EXCEPTION); 
  			MYRAISEEXCEPTION;
		}
		
		//Release from FindObject
		pDispatch->Release();
		pDispatch = NULL;
	
	    return; 
	}

	EXCEPT{
		if(pDispatch)
			pDispatch->Release();
						
		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::Remove")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return;
	}
}

/*
 * CDataObjects::CreateObject
 *
 * Purpose:
 *  Contains same functionality as VBA's CreateObject. VBA developer
 *	must use this CreateObject to use add-in
 *
 * Parameters
 *	bstrProgID	ProgID of WOSA/XRT object to instantiate
 *	vName		Name of object to create
 *  vRange		Range to recalc for TypeB updating (Optional)
 *
 * Return Value
 *	LPDISPATCH
 *
 */
  

STDMETHODIMP_(IDispatch *) 
CDataObjects::CreateObject(BSTR bstrProgID, VARIANT vName, VARIANT vRange)
{                                       
	CLSID clsid;
	LPDISPATCH pXRTDispatch = NULL;        
    CDataObject *pDataObject = NULL;

	TRY{
		// Is name valid    
		if(FAILED(CheckPointer((LPVOID) bstrProgID, 0))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::CreateObject")
					, IDS_DEBUG_INVALIDNAME
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}
	
		if(FAILED(CheckVariant(&vName, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::CreateObject")
					, IDS_DEBUG_INVALIDNAME
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}
	
		// Make sure we do not have one by this name already
		if(SUCCEEDED(m_col.FindObject(vName.bstrVal, NULL, NULL))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::CreateObject")
					, IDS_ALREADYINCOLLECTION
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}

		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 0.0, TRUE);
	    
	    // get CLSID of ProgID
    	if(FAILED(CLSIDFromProgID(bstrProgID, &clsid))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::CreateObject")
					, IDS_DEBUG_CLSIDFROMPROGIDFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}

		// Create new object
    	if(FAILED(CoCreateInstance(clsid, NULL, CLSCTX_LOCAL_SERVER
    			, IID_IDispatch, (LPVOID FAR *)  &pXRTDispatch))){
			
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::CreateObject")
					, IDS_DEBUG_COCREATEINSTANCEFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}
                                          
		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 1.0, TRUE);
		HideStatusWindow();

		//DO NOT RELEASE pXRTDispatch because DataObject will become responsible for it	
		
		// Now add it as any object
		return Add(pXRTDispatch, vName, vRange);
	}

	EXCEPT{
		HideStatusWindow();

		if(pXRTDispatch) 
			pXRTDispatch->Release();
		
	    m_bRaiseException = TRUE;
	    
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::CreateObject")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;
	}
} 


/*
 * CDataObjects::GetObject
 *
 * Purpose:
 *  Contains same functionality as VBA's GetObject. VBA developer
 *	must use this GetObject to use add-in
 *
 * Parameters
 *	FileName	FileName of object to load
 *	Moniker		Moniker for more specific binding
 *  Range		Range to recalc for TypeB updating (Optional)
 *
 * Return Value
 *	LPDISPATCH
 *
 */

STDMETHODIMP_(IDispatch *)
CDataObjects::GetObject(BSTR FileName, VARIANT Moniker, VARIANT Range)
{
	VARIANT vName;
	LPDISPATCH pXRTDispatch = NULL;
	LPDISPATCH pNewDispatch = NULL;
	CDataObject FAR *pNewDataObject = NULL;
	STGMEDIUM stg;

	TRY{
		if(FAILED(CheckPointer((LPVOID) FileName, 0))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_INVALIDNAME
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}

        // GetObject can be used in three different ways depending on the params
		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 0.0, TRUE);

    	// If FileName is valid and Moniker is NULL	
        if(FileName != NULL && Moniker.vt == VT_ERROR) {
            // Have object load file
    		if(FAILED(XRTLoadPersistFile(FileName, &pXRTDispatch))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_IPERSISTFILELOADFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
    		}
    		
        // Else if FileName and Moniker is specified
        }else if(FileName != NULL && Moniker.vt != VT_ERROR){
        	LPBC pBC = NULL;
        
			if(FAILED(CreateBindCtx(0L, &pBC))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_CREATEBINDCTXFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}

			DWORD dwChars;
			LPMONIKER pMoniker;
			
			if(FAILED(MkParseDisplayName(pBC, Moniker.bstrVal, &dwChars,
					&pMoniker))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_MKPARSEDISPLAYNAMEFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}					
			
			if(FAILED(pMoniker->BindToObject(pBC, NULL, IID_IDispatch
					, (LPVOID FAR *) &pXRTDispatch))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_BIDTOOBJECTFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}
			
			pMoniker->Release();
			pMoniker = NULL;
			
			pBC->Release();
			pBC = NULL;	
				
		// Else if filename is NULL and Moniker is valid		        
		}else if(FileName == NULL && Moniker.vt == VT_ERROR){
			CLSID clsid;
			LPUNKNOWN pUnk = NULL;
		
			if(FAILED(CLSIDFromProgID(Moniker.bstrVal, &clsid))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_CLSIDFROMPROGIDFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}
			
			if(FAILED(GetActiveObject(clsid, NULL, (LPUNKNOWN FAR *) &pUnk))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_GETACTIVEOBJECTFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}
			
			if(FAILED(pUnk->QueryInterface(IID_IDispatch, (LPVOID FAR *) pXRTDispatch))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::GetObject")
					, IDS_DEBUG_QIFORIDISPATCHFAILED
					, DISP_E_EXCEPTION); 
				MYRAISEEXCEPTION;
			}
		}else{             
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::GetObject")
				, IDS_DEBUG_INVALIDPARAMETERS
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}
		        
		VariantInit(&vName);
		vName.vt = VT_BSTR;
    	vName.bstrVal = SysAllocString(TEXT("DataObject1"));
                                          
		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 0.0, TRUE);
		DisplayStatusWindow(TEXT("Updating add-in's collection..."), (float) 0.0, TRUE);
	
		pNewDispatch = Add(pXRTDispatch, vName, Range);

   		if(pNewDispatch == NULL){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::GetObject")
				, IDS_DEBUG_DATAOBJSADDFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
   		}
        
	    // Internal QI; just a recast
		pNewDispatch->QueryInterface(IID_IMyDataObject, (LPVOID FAR *) &pNewDataObject);    

		pNewDataObject->m_bDontIncludeNames = TRUE;
    
		DisplayStatusWindow(TEXT("Updating add-in's collection..."), (float) 1.0, FALSE);
		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 0.0, TRUE);
	
    	// GetData to fill cache
    	if(FAILED(pNewDataObject->m_pXRTDataObject->GetData(
    			&pNewDataObject->m_fte, &stg))){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::GetObject")
				, IDS_DEBUG_IDATAOBJGETDATAFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}            

		DisplayStatusWindow(TEXT("Loading Real-Time Data Object..."), (float) 1.0, TRUE);
		DisplayStatusWindow(TEXT("Intializing add-in's cache..."), (float) 0.0, TRUE);

		// Create a cache    	
    	if(FAILED(CCache::Create(pNewDataObject, FALSE
    			, &pNewDataObject->m_pCache) )){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::GetObject")
				, IDS_DEBUG_CACHECREATEFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
        }
	
		// Fill the cache        
    	if(FAILED(pNewDataObject->m_pCache->Fill(&stg))){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::GetObject")
				, IDS_DEBUG_CACHEFILLFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}


		#ifdef WIN32
	    if(FAILED(CDataObject::RetrieveDataObjectName(pXRTDispatch, &vName.bstrVal))){
    		m_excepinfo.scode = E_FAIL;
    		m_excepinfo.bstrDescription = SysAllocString(TEXT("Could not retrieve name from XRT object!"));
			MYRAISEEXCEPTION;
    	}
		#endif
		
    	pNewDataObject->m_bstrName = SysAllocString(vName.bstrVal); 
		
    	pNewDataObject->Release(); // Release from QI

		VariantClear(&vName);                                    
		
		ReleaseStgMedium(&stg);
	
		HideStatusWindow();
 
		return pNewDispatch;
	}

	EXCEPT{
		HideStatusWindow();

    	m_bRaiseException = TRUE;    
	
		VariantClear(&vName);
		
		ReleaseStgMedium(&stg);
	    
	    if(pNewDataObject)
			delete pNewDataObject;
		
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::GetObject")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	
		return NULL;
	}
}

 
STDMETHODIMP_(LPDISPATCH)
CDataObjects::get_Item(VARIANT Index)  
{   
	LPDISPATCH pDispatch;
	HRESULT hr;

	TRY{
		if(FAILED(CheckVariant(&Index, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObjects::get_Item")
					, IDS_DEBUG_INVALIDNAME
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}
		
		// check index    
   		if(Index.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) Index.iVal, &pDispatch);
   		}else if(Index.vt == VT_I4){
	   		hr = m_col.FindObjectByIndex(Index.lVal, &pDispatch);
   		}else if(Index.vt == VT_BSTR){
	   		hr = m_col.FindObject(Index.bstrVal, &pDispatch, NULL);
   		}else{
			AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObjects::get_Item")
				, IDS_DEBUG_FINDOBJECTFAILED
				, DISP_E_EXCEPTION); 
  			MYRAISEEXCEPTION;
	   	}

		VariantClear(&Index);
	
    	return pDispatch;
	}
	
	EXCEPT{
		m_bRaiseException = TRUE;
		
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::get_Item")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;
	}
}

/*
 * CDataObjects::get_Count
 *
 * Purpose:
 *  Returns number of items in collection.
 *
 */
STDMETHODIMP_(long)     
CDataObjects::get_Count(void)  
{
    return m_col.m_cElements;             
}

 
/*
 * CDataObjects::get_NewEnum
 *
 * Purpose:
 *  Returns an enumerator (IEnumVARIANT) for the items curently in the collection. 
 *  The NewEnum property is restricted and so is invisible to users of an 
 *  automation controller's scripting language. Automation controllers that support
 *  a 'For Each' statement to iterate through the elements of a collection will use
 *  the enumerator returned by NewEnum. The enumerator creates a snapshot of the
 *  the current state of the collection.
 *
 */
STDMETHODIMP_(LPUNKNOWN)
CDataObjects::get__NewEnum(void)
{
    LPUNKNOWN punkEnumVariant = NULL;

	TRY{
		if(FAILED(m_col.GetEnum(&punkEnumVariant))){
			MYRAISEEXCEPTION;	
		}
		
    	return punkEnumVariant; 
	}
    
	EXCEPT{
    	m_bRaiseException = TRUE;    
		
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::_NewEnum")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;    	
	}
}


HRESULT CDataObjects::DisplayStatusWindow(BSTR bstrMsg, float fComplete
		, BOOL bInvalidate)
{	
	if(!m_pRealTimeExt->m_bEnableStatusWindow)
		return S_OK;

	if(m_bstrStatusWndMsg == NULL 
			|| (bstrMsg != NULL && lstrcmp(bstrMsg, m_bstrStatusWndMsg))){
		
		SysReAllocString(&m_bstrStatusWndMsg, bstrMsg);
	}

	m_fComplete = fComplete;	
	
	if(!IsWindowVisible(m_hWnd)){
		ShowWindow(m_hWnd, SW_SHOWNORMAL);
		bInvalidate = TRUE;
	}

	InvalidateRect(m_hWnd, NULL, bInvalidate);
	UpdateWindow(m_hWnd);

	return S_OK;
}

HRESULT CDataObjects::HideStatusWindow()
{
	if(IsWindowVisible(m_hWnd)){
		ShowWindow(m_hWnd, SW_HIDE);
	}

	return S_OK;
}
