// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Spreadsheet callable function which returns a specified value to 
//		the cell containing the function.
//	

#include "precomp.h"
#include "framewrk.h"
#include "realtime.h"
#include "dataobjs.h"
#include "dataobj.h"
#include "props.h"
#include "prop.h"
#include "cache.h"
#include "guid.h"
#include "fn.h"

/*
 * REALTIMEUPDATE
 * XLS CALLABLE FUNCTION
 *
 * Purpose:
 *  The XLS calls this function from within any cell. Returns the value
 *	of a specific proeprty within a DataObject.
 *
 * Parameters:
 *  pxDataObject    Name or reference to cell containg the name of the DataObject
 *  pxItem			Name or reference to cell containg the name of the Request
 *  pxProperty		Name or reference to cell containg the name of the Property
 *
 * Return Value:
 *  HRESULT
 *
 */

  
extern "C"
LPXLOPER EXPORT FAR PASCAL REALTIMEUPDATE(LPXLOPER pxDataObject
		, LPXLOPER pxItem, LPXLOPER pxProperty )
{                          
	XLOPER x;                            
	LPXLOPER pX = &x;

	CDataObject FAR *pDataObject = 0;
	
	XLOPER xItemResult;
	XLOPER xPropResult;
	XLOPER xObjectResult;
    
    int nCount;
 	
 	TCHAR szProperty[MAX_STRING_SIZE + 1];
 	TCHAR szItem[MAX_STRING_SIZE + 1];
 	TCHAR szObjectName[MAX_STRING_SIZE + 1];

	memset(szObjectName, 0, sizeof(szObjectName));
	memset(szItem, 0, sizeof(szItem));
	memset(szProperty, 0, sizeof(szProperty));

    LPTSTR pItemName = szItem;
    LPTSTR pPropertyName = szProperty;
    
    XLOPER FAR *pxlPropertyValue;      
    
    LPDISPATCH pDispatch = NULL;

	TRY{        
		// My only significant global var
		if(g_pRealTimeExt == NULL){	             
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_OBJECTNOTINITAILIZED
				, NULL); 
			MYRAISEEXCEPTION;
    	}
          
		// Get DataObject name
		if(pxDataObject->xltype == xltypeStr){    //STRING
			nCount = 0;
			nCount = lstrlenA((LPSTR) (pxDataObject->val.str + sizeof(char)));
			nCount = (int)*(pxDataObject->val.str);		                

			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (pxDataObject->val.str + sizeof(char))
						, nCount, szObjectName, 256);
			#else 
				lstrcpyn(szObjectName, (LPTSTR) (pxDataObject->val.str + sizeof(TCHAR)), nCount+1); 
			#endif
		}else if(pxDataObject->xltype == xltypeRef){   //CELL REFERENCE
			long x;
		
			x = (long) Excel(xlCoerce, (LPXLOPER) &xObjectResult, 2, (LPXLOPER) pxDataObject, (LPXLOPER) TempInt(xltypeStr));
		              
			if(xObjectResult.xltype != xltypeStr){
				AddInDebugOutput(NULL
					, TEXT("REALTIMEUPDATE")
					, IDS_DEBUG_DATAINCELLREFENCEISNOTVALID
					, NULL); 
				MYRAISEEXCEPTION;
			}
		
			nCount = 0;
			nCount = lstrlenA((LPSTR)(xObjectResult.val.str + sizeof(char)));           
			nCount = (int)*(xObjectResult.val.str);		                
	
			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (xObjectResult.val.str + sizeof(char))
						, nCount, szObjectName, 256);
			#else 
				lstrcpyn(szObjectName, (LPTSTR) (xObjectResult.val.str + sizeof(TCHAR)), nCount+1); 
			#endif

		}else{
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_INVALIDDATAOBJECTNAME
				, NULL); 
			MYRAISEEXCEPTION;
		}
	
		//Find the DataObject
		if(FAILED(g_pRealTimeExt->m_pDataObjects->m_col.FindObject(szObjectName, &pDispatch, NULL))){
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_DATAOBJNOTFOUNDINCOLLECTION
				, NULL); 
			MYRAISEEXCEPTION;
		}
	    
	    // Get IDataObject by recasting
	    // Internal QI is just a recast
	    pDispatch->QueryInterface(IID_IMyDataObject, (LPVOID FAR*) &pDataObject);
		pDispatch->Release();
	    pDispatch = NULL;
	           
		// Get Request name
		if(pxItem->xltype == xltypeStr){ // STRING
			nCount = 0;
			nCount = lstrlenA((LPSTR) (pxItem->val.str + sizeof(char)));           			
			nCount = (int)*(pxItem->val.str);		                
	
	    	// Can't rely on XL's allocations
			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (pxItem->val.str + sizeof(char))
						, nCount, szItem, 256);
			#else 
				lstrcpyn(szItem, (LPTSTR) (pxItem->val.str + sizeof(TCHAR)), nCount+1); 
			#endif
		}else if(pxItem->xltype == xltypeRef){ //CELL REFERENCE
			long x;
		
			x = (long) Excel(xlCoerce, (LPXLOPER) &xItemResult, 2, (LPXLOPER) pxItem, (LPXLOPER) TempInt(xltypeStr));
		              
			if(xItemResult.xltype != xltypeStr){
				AddInDebugOutput(NULL
					, TEXT("REALTIMEUPDATE")
					, IDS_DEBUG_DATAINCELLREFENCEISNOTVALID
					, NULL); 
				MYRAISEEXCEPTION;
			}
		
			nCount = 0;
			nCount = (int)*(xItemResult.val.str);		                
	
	    	// Can't rely on XL's allocations
			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (xItemResult.val.str + sizeof(char))
						, nCount, szItem, 256);
			#else 
				lstrcpyn(szItem, (LPTSTR) (xItemResult.val.str + sizeof(TCHAR)), nCount+1); 
			#endif

		}else{
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_INVALIDREQUESTNAME
				, NULL); 
			MYRAISEEXCEPTION;
		}

		// 	Get Property name
		if(pxProperty->xltype == xltypeStr){ //STRING
			nCount = 0;
			nCount = lstrlenA((LPSTR) (pxProperty->val.str + sizeof(char)));
			nCount = (int)*(pxProperty->val.str);		                
	    
	    	// Can't rely on XL's allocations
			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (pxProperty->val.str + sizeof(char))
						, nCount, szProperty, 256);
			#else 
				lstrcpyn(szProperty, (LPTSTR) (pxProperty->val.str + sizeof(TCHAR)), nCount+1); 
			#endif
		}else if(pxProperty->xltype == xltypeRef){   // CELL REFERENCE
			Excel(xlCoerce, (LPXLOPER) &xPropResult, 2, (LPXLOPER) pxProperty, (LPXLOPER) TempInt(xltypeStr));
		
			if(xPropResult.xltype != xltypeStr){
				AddInDebugOutput(NULL
					, TEXT("REALTIMEUPDATE")
					, IDS_DEBUG_DATAINCELLREFENCEISNOTVALID
					, NULL); 
				MYRAISEEXCEPTION;
			}
		
			nCount = 0;
			nCount = lstrlenA((LPSTR) (xPropResult.val.str + sizeof(char)));
			nCount = (int)*(xPropResult.val.str);		                
	
	    	// Can't rely on XL's allocations
			#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, NULL
						, (LPSTR) (xPropResult.val.str + sizeof(char))
						, nCount, szProperty, 256);
			#else 
				lstrcpyn(szProperty, (LPTSTR) (xPropResult.val.str + sizeof(TCHAR)), nCount+1); 
			#endif

		}else{
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_INVALIDPROPNAME
				, NULL); 
			MYRAISEEXCEPTION;
		}
	    
	    //
	    //	Now I finally have the criteria so...
	    //
	    
		if(FAILED( pDataObject->m_pCache->GetPropertyValue
				(pItemName, pPropertyName, &pxlPropertyValue))){
			AddInDebugOutput(NULL
				, TEXT("REALTIMEUPDATE")
				, IDS_DEBUG_COULDNOTFINPROPERTY
				, NULL); 
			MYRAISEEXCEPTION;
	 	}
	
		// Clear DataObject if it was passed as cell reference
		if(pxDataObject->xltype == xltypeRef){   // CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xObjectResult);
    	}      
	
		// Clear Request if it was passed as cell reference
		if(pxItem->xltype == xltypeRef){ //CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xItemResult);
		}
	
		// Clear Property if it was passed as cell reference
		if(pxProperty->xltype == xltypeRef){   // CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xPropResult);
    	}

		// Release from FindDataObject    
		pDataObject->Release();
		pDataObject = NULL;
        
        return pxlPropertyValue;
	}

	EXCEPT{
		if(pDispatch)
			pDispatch->Release();
			
		if(pDataObject)
			pDataObject->Release();
	
		if(pxItem->xltype == 0x400){ //CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xItemResult);
		}
	
		if(pxProperty->xltype == 0x400){   // CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xPropResult);
    	}
    
		if(pxDataObject->xltype == 0x400){   // CELL REFERENCE
			Excel(xlFree, 0, 1,(LPXLOPER) &xObjectResult);
    	}

		pX->xltype = xlerrNull;
		pX->val.err = xlerrValue;

   		AddInDebugOutput(NULL, TEXT("REALTIMEUPDATE")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return pX;
	}
}                                                       

