// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Implementation of IAdviseSink. This enables add-in to receive asynch. 
//		data from realtime data object.

#include "precomp.h"
#include "iadvsink.h"
#include "realtime.h"
#include "dataobj.h"
#include "dataobjs.h"
#include "cache.h"

/*
 * CImpAdviseSink::CImpAdviseSink
 *
 * Purpose:
 *  Constructor
 *
 */ 

CImpIAdviseSink::CImpIAdviseSink(void)
{                                                            
    return;
}

/*
 * CImpAdviseSink::~CImpAdviseSink
 *
 * Purpose:
 *  Destructor
 *
 */ 

CImpIAdviseSink::~CImpIAdviseSink(void)
{
    return;
}

/*
 * CImpIAdviseSink::QueryInterface
 * CImpIAdviseSink::AddRef
 * CImpIAdviseSink::Release
 *
 * Purpose:
 *  IUnknown members for CImpIAdviseSink object.
 */

STDMETHODIMP CImpIAdviseSink::QueryInterface(REFIID riid, LPVOID FAR *ppv)
{
    *ppv=NULL;

    //Any interface on this object is the object pointer.
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IAdviseSink))
        *ppv=(LPVOID)this;

    /*
     * If we actually assign an interface to ppv we need to AddRef it
     * since we're returning a new pointer.
     */
     
    if (NULL!=*ppv){
        ((LPUNKNOWN)*ppv)->AddRef();
        return S_OK;
    }

    return ResultFromScode(E_NOINTERFACE);
}


STDMETHODIMP_(ULONG) CImpIAdviseSink::AddRef(void)
{
    return ++m_cRef;
}


STDMETHODIMP_(ULONG) CImpIAdviseSink::Release(void)
{
    if (--m_cRef == 0){
        delete this;
    }

    return m_cRef;
}


/*
 * IAdviseSink::OnDataChange
 *
 * Purpose:
 *  Notifes the advise sink that data changed in a data object.  On
 *  this message you may request a new data rendering and update your
 *  displays as necessary.  Any data sent to this function is owned
 *  by the caller, not by this advise sink!
 *
 *  All Advise Sink methods should be considered asynchronous and therefore
 *  we should attempt no synchronous calls from within them to an EXE
 *  object.  If we do, we'll get RPC_E_CALLREJECTED as shown below.
 *
 * Parameters:
 *  pFE             LPFORMATETC describing format that changed
 *  pSTM            LPSTGMEDIUM providing the medium in which the data
 *                  is provided.
 *
 * Return Value:
 *  None
 */

STDMETHODIMP_(void) CImpIAdviseSink::OnDataChange(LPFORMATETC pFE, LPSTGMEDIUM pSTM)
{
	TRY{
		//Check to see if its the right aspect           
    	if (!(DVASPECT_CONTENT & pFE->dwAspect))
        	return;
                                  
		//Check to see if its the right format
    	if (pFE->cfFormat == g_cfWOSAXRT && pSTM->tymed & TYMED_HGLOBAL){
			if(!m_pDataObject){
				AddInDebugOutput(NULL
					, TEXT("IAdviseSink::OnDataChange")
					, IDS_DEBUG_INAVALIDCDATAOBJPTR
					, NULL); 
				MYRAISEEXCEPTION;
			}		
	
			#ifdef _WIN32
				// Update cache
				if(FAILED(m_pDataObject->m_pCache->Update(pSTM))){
					AddInDebugOutput(NULL
						, TEXT("IAdviseSink::OnDataChange")
						, IDS_DEBUG_CCACHEUPDATEFAILED
						, NULL); 
					MYRAISEEXCEPTION;
				}                       
			
				// If we receive a valid range, recalc that range			
				if(m_pDataObject->m_pUpdateRangeDispatch){
					XLCalculateNow(m_pDataObject);
				}

				// If we received a valid basci callback func.
				// call that name.		
				if(m_pDataObject->m_bstrXLCallBackFunc != NULL) {
					XLRunMacro(m_pDataObject);
				}
			#else
 				#ifdef WIN16HARDUPDATES
		    		SendMessage(m_pDataObject->m_hWnd, g_msgXRTDBUpdate, 0, (LPARAM) pSTM);
		    	#else
					MSG msg;

		    		PostMessage(m_pDataObject->m_hWnd, g_msgXRTDBUpdate, 0, (LPARAM) pSTM);

    				while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)){
    					TranslateMessage(&msg);
    					DispatchMessage(&msg);
    				}
				#endif
    			
			#endif
		}
		
		return;                       
	}

	EXCEPT{
		ReleaseStgMedium( pSTM );

   		AddInDebugOutput(NULL, TEXT("IAdviseSink::OnDataChange")
   			, IDS_DEBUG_EXCEPTION, NULL); 

    	return;
    }
}

/*
 * IAdviseSink::OnClose
 *
 * Purpose:
 *  Informs the advise sink that the OLE object has closed and is
 *  no longer bound in any way.  On this you typically change state
 *  variables and redraw shading, etc.
 *
 * Parameters:
 *  None
 *
 * Return Value:
 *  None
 */

STDMETHODIMP_(void) CImpIAdviseSink::OnClose(void)
{
	VARIANT v;
	
	if(!m_pDataObject){
		AddInDebugOutput(NULL
			, TEXT("IAdviseSink::OnDataChange")
			, IDS_DEBUG_INAVALIDCDATAOBJPTR
			, NULL); 
		return;
	}		
	
	VariantInit(&v);
	v.vt = VT_BSTR;
	v.bstrVal = SysAllocString(m_pDataObject->m_bstrName);
	
	m_pDataObject->m_dwAdviseConnection	= 0;
	m_pDataObject->m_pRealTimeExt->m_pDataObjects->Remove(v);
	
	VariantClear(&v);            
	
    return;
}
        
/*
 * IAdviseSink::OnViewChange
 *
 * Purpose:
 *  Notifes the advise sink that presentation data changed in the data
 *  object to which we're connected providing the right time to update
 *  displays using such presentations.
 *
 * Parameters:
 *  dwAspect        DWORD indicating which aspect has changed.
 *  lindex          LONG indicating the piece that changed.
 *
 * Return Value:
 *  None
 */

STDMETHODIMP_(void) CImpIAdviseSink::OnViewChange(DWORD dwAspect, LONG lindex)
    {
    return;
    }





/*
 * IAdviseSink::OnRename
 *
 * Purpose:
 *  Informs the advise sink that an IOleObject has been renamed, primarily
 *  when its linked.
 *
 * Parameters:
 *  pmk             LPMONIKER providing the new name of the object
 *
 * Return Value:
 *  None
 */

STDMETHODIMP_(void) CImpIAdviseSink::OnRename(LPMONIKER pmk)
    {
    return;
    }






/*
 * IAdviseSink::OnSave
 *
 * Purpose:
 *  Informs the advise sink that the OLE object has been saved
 *  persistently.  The primary purpose of this is for containers that
 *  want to make optimizations for objects that are not in a saved
 *  state, so on this you have to disable such optimizations.
 *
 * Parameters:
 *  None
 *
 * Return Value:
 *  None
 */

STDMETHODIMP_(void) CImpIAdviseSink::OnSave(void)
    {
    return;
    }





