// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CRealTimeExt class implements the IUnknown and IDispatch interfaces.
// 		The properties and methods are unique to to add-in
// 		and are defined throughout this module... 

#include "precomp.h"

#define INITGUID	

#include <initguid.h>
#include "guid.h"
#include "realtime.h"
#include "dataobj.h"
#include "dataobjs.h"

#include "framewrk.h"
/*
 * CRealTimeExt::Create
 *
 * Purpose:
 *  Creates an instance of the client object and initializes it.
 *
 * Parameters:
 *  ppRealTimeExt    Returns Item object.
 *
 * Return Value:
 *  HRESULT
 *
 */
 
HRESULT CRealTimeExt::Create(CRealTimeExt FAR* FAR* ppRealTimeExt ) 
{   
    CRealTimeExt FAR* pRealTimeExt;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppRealTimeExt, 0))){
    		AddInDebugOutput(NULL, TEXT("CRealTimeExt::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
	    *ppRealTimeExt = NULL;

		// Create real-time obejct
		pRealTimeExt = new CRealTimeExt();
		
		if(pRealTimeExt == NULL){
			AddInDebugOutput(&pRealTimeExt->m_excepinfo
					, TEXT("CRealTimeExt::Create")
					, IDS_DEBUG_REALTIMEALLOCFAILED
					, NULL); 
		}
						
		// Load type library for future use    	
    	if (FAILED(LoadTypeInfo(&pRealTimeExt->m_ptinfo, IID_IRealTimeExt))){
			AddInDebugOutput(&pRealTimeExt->m_excepinfo
					, TEXT("CRealTimeExt::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
    	}

        // Create DataObjects
	    if (FAILED(CDataObjects::Create((CDataObjects FAR* FAR* ) &pRealTimeExt->m_pDataObjects))){
			AddInDebugOutput(&pRealTimeExt->m_excepinfo
					, TEXT("CRealTimeExt::Create")
					, IDS_DEBUG_CREATEREALTIMEEXTFAILED
					, NULL); 
			MYRAISEEXCEPTION;
	    }

		// Save pointer to realtime in DataObjects
    	pRealTimeExt->m_pDataObjects->m_pRealTimeExt = pRealTimeExt;
    	pRealTimeExt->m_pDataObjects->AddRef();
    	
    	// Set name	    
		pRealTimeExt->m_bstrName = SysAllocString(g_szTitle);
        
    	*ppRealTimeExt = pRealTimeExt;       
		
    	return S_OK;
	}    

	EXCEPT{
	    if (pRealTimeExt == NULL)
    	    return ResultFromScode(E_FAIL);
        
   		delete pRealTimeExt;

   		AddInDebugOutput(NULL , TEXT("CRealTimeExt::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 

	    return ResultFromScode(E_FAIL);
	}
}

/*
 * CRealTimeExt::CRealTimeExt
 *
 * Purpose:
 *  Constructor for CRealTimeExt object. Initializes members to NULL.
 *
 */                             
 
CRealTimeExt::CRealTimeExt()
{
	m_pXLAppDispatch = NULL;
	m_pDataObjects = NULL;
	m_bstrName = NULL;
	m_cRef = 1;
	m_bEnableStatusWindow = TRUE;
    m_bRaiseException = FALSE;
}

/*
 * CRealTimeExt::~CRealTimeExt
 *
 * Purpose:
 *  Destructor for CRealTimeExt object. 
 *
 */

CRealTimeExt::~CRealTimeExt()
{                 

	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);
		
	if(SUCCEEDED(CheckPointer(m_pDataObjects)))
		delete m_pDataObjects;
		
	if(SUCCEEDED(CheckPointer(m_ptinfo)))
       	m_ptinfo->Release();

	if(SUCCEEDED(CheckDispatch(m_pXLAppDispatch)))
		m_pXLAppDispatch->Release();

    // Free BSTR IMalloc
	BSTRToXLOPER(NULL, NULL, TRUE);

	g_cObjCount--;    
}

/*
 * CRealTimeExt::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CRealTimeExt::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IRealTimeExt) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DRealTimeExt)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CRealTimeExt::AddRef(void)
{ 
    return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CRealTimeExt::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CRealTimeExt::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */     
 
STDMETHODIMP
CRealTimeExt::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CRealTimeExt::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */
 
STDMETHODIMP
CRealTimeExt::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CRealTimeExt::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */

STDMETHODIMP 
CRealTimeExt::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CRealTimeExt::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *
 */
STDMETHODIMP
CRealTimeExt::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

    m_bRaiseException = FALSE;
	 
    hr = DispInvoke(
		this, m_ptinfo,
		dispidMember, wFlags, pdispparams,
		pvarResult, pexcepinfo, puArgErr);
    
    // Check if property accessor function or method called by DispInvoke wants to raise
    // an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
    // structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
    // property or method.    
    if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);
		
		m_bRaiseException = FALSE;
        m_excepinfo.scode = NULL;
        m_excepinfo.wCode = 0;
        
		return ResultFromScode(DISP_E_EXCEPTION); 
    }else{
    	 return hr;    
	}
}

/*
 * CRealTimeExt::Quit
 *
 * Purpose:
 *  Quits XLL
 *
 */                                 
 
STDMETHODIMP_(void)     
CRealTimeExt::Quit(void)  
{           
	delete this;

	return;	                
}

/*
 * CRealTimeExt::get_Name
 *
 * Purpose:
 *  Returns a realtime object name
 *
 */                                 
 
STDMETHODIMP_(BSTR)     
CRealTimeExt::get_Name(void)  
{
    return SysAllocString(m_bstrName);
}

 
/*
 * CRealTimeExt::About
 *
 * Purpose:
 *  Displays Add-In's about box
 *
 */                                 
 
STDMETHODIMP_(void)     
CRealTimeExt::About(void)  
{                                                                                          
	TCHAR buf[256];

	memset(buf, (char) 0, sizeof(buf));

	lstrcpy(buf, TEXT("Excel Real-Time Extensions\n\rVersion 1.0.0.\r\n\r\n"));

	lstrcat(buf, TEXT("Copyright "));

	buf[lstrlen(buf)] = (TCHAR) (BYTE) 169;

	lstrcat(buf, TEXT("1994 Microsoft Corporation.\r\nAll Rights Reserved.\r\n\r\n"));
	lstrcat(buf, TEXT("By Rich Ciapala - Systems Engineer\r\n"));
	lstrcat(buf, TEXT("Microsoft New York/New Jersey\r\n"));

	XLOPER xDLL;

    Excel( xlGetHwnd, &xDLL, 0 ) ;
   	HWND hwndXL = (HWND)xDLL.val.w ;
	
	MessageBox(hwndXL, buf, g_szTitle, MB_OK);

    return;             
}

/*
 * CRealTimeExt::DataObjects
 *
 * Purpose:
 *  Returns DataObjects
 *
 */                                 
 
STDMETHODIMP_(IDispatch *)
CRealTimeExt::get_DataObjects(void)
{                               
	LPDISPATCH pDispatch;
	
	//QI will do AddRef
	HRESULT hr = m_pDataObjects->QueryInterface(IID_IDispatch, (LPVOID FAR *)&pDispatch);
	
	if(SUCCEEDED(hr)){
		return pDispatch;
	}else{
   		AddInDebugOutput(NULL , TEXT("CRealTimeExt::get_DataObjects")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		m_bRaiseException = TRUE;	
   		return NULL;
    }
}

                                
STDMETHODIMP_(void)
CRealTimeExt::put_EnableStatusWindow(short Value)
{
	m_bEnableStatusWindow = Value;
}

                                
STDMETHODIMP_(short)
CRealTimeExt::get_EnableStatusWindow()
{
	return m_bEnableStatusWindow;
}

/*
 * CRealTimeExt::XLApplication
 * Property Write-Only
 *
 * Purpose:
 *  Required to call OnDataChange function
 *
 */                                 
 
STDMETHODIMP_(void) 
CRealTimeExt::put_XLApplication(IDispatch * XLAppObject){
	LPDISPATCH lpDispatch = NULL;

	if(FAILED(CheckDispatch(XLAppObject))){
   		AddInDebugOutput(NULL , TEXT("CRealTimeExt::put_XLApplication")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		m_bRaiseException = TRUE;	

   		return;
	}

	if(m_pXLAppDispatch){
		m_pXLAppDispatch->Release();	
	}
		
	m_pXLAppDispatch = XLAppObject;
	
	m_pXLAppDispatch->AddRef();
	
	return;		
}

