// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CRequests class implements the IUnknown and IDispatch interfaces.
// 		The IDispatch interface supersets all the WOSA/XRT Requests methods 
// 		and properties.


#include "precomp.h"
#include "guid.h"

#include "dataobj.h"
#include "requests.h"
#include "request.h"
#include "props.h"
#include "prop.h"
#include "enumvar.h"
#include "cache.h"

/*
 * CRequests::Create
 *
 * Purpose:
 *  Creates an instance of a collection object and initializes it.
 *
 * Parameters:
 *  pXRTDispatch  Requests associted XRT object dispatch
 *  ppRequests    Returns collection object.
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CRequests::Create(LPDISPATCH pXRTDispatch
		, CRequests FAR* FAR* ppRequests) 
{   
    CRequests FAR* pRequests = NULL;
	     
	TRY{ 
		if(FAILED(CheckDispatch(pXRTDispatch))){
    		AddInDebugOutput(NULL, TEXT("CCache::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckPointer((LPVOID) ppRequests, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	   
	    *ppRequests = NULL;
    
    	// Create new collection
    	pRequests = new CRequests();
	    if (pRequests == NULL){
			AddInDebugOutput(&pRequests->m_excepinfo
					, TEXT("CRequests::Create")
					, IDS_DEBUG_REQUESTSALLOCFAILED
					, NULL); 
			return ResultFromScode(E_FAIL);
    	}                                                                         
    
	    // Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
    	if (FAILED(LoadTypeInfo(&pRequests->m_ptinfo, IID_IRequests))){
			AddInDebugOutput(&pRequests->m_excepinfo
					, TEXT("CRequests::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
    	}
    
		//Save parallel XRT Reests
    	pRequests->m_pXRTDispatch = pXRTDispatch;
    	
    	pRequests->m_bRaiseException = FALSE;

    	*ppRequests = pRequests;

    	return S_OK;
	}
    
	EXCEPT{ 
    	if (pRequests == NULL)
			return ResultFromScode(E_OUTOFMEMORY);

   		AddInDebugOutput(NULL, TEXT("CRequests::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 
        
        pRequests->m_bRaiseException = TRUE;

    	delete pRequests;
        
		return ResultFromScode(E_FAIL);
	}
}

/*
 * CRequests::CRequests
 *
 * Purpose:
 *  Constructor for CRequests object. Initializes members to NULL.
 *
 */

CRequests::CRequests()
{    
	m_pXRTDispatch = NULL;						 
	m_pDataObject = NULL;
	m_dwMaxColCount = 0;
	m_bstrName = NULL;
	m_ptinfo = NULL;
	m_cRef = 1;
   	m_bRaiseException = 0;
}

/*
 * CRequests::~CRequests
 *
 * Purpose:
 *  Destructor for CRequests object. 
 *
 */

CRequests::~CRequests()
{                     
	if(SUCCEEDED(CheckPointer(m_ptinfo)))
   		m_ptinfo->Release();         
		
	if(SUCCEEDED(CheckDispatch(m_pXRTDispatch)))
		m_pXRTDispatch->Release();		
		
	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);

}


/*
 * CRequests::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CRequests::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IRequests) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DRequests)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CRequests::AddRef(void)
{ 
    return --m_cRef;
}


STDMETHODIMP_(ULONG)
CRequests::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CRequests::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */
STDMETHODIMP
CRequests::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CRequests::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */
STDMETHODIMP
CRequests::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CRequests::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */
STDMETHODIMP 
CRequests::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CRequests::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 */
STDMETHODIMP
CRequests::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

    m_bRaiseException = FALSE;
	 
    hr = DispInvoke(          
		this, m_ptinfo,
		dispidMember, wFlags, pdispparams,
		pvarResult, pexcepinfo, puArgErr);
    
    // Check if property accessor function or method called by DispInvoke wants to raise
    // an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
    // structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
    // property or method.    
    if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);
		
		m_bRaiseException = FALSE;
        m_excepinfo.scode = NULL;
        m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
    }else{
    	return hr;    
	}
}


/*
 * CRequests::Add
 *
 * Purpose:
 *  Adds an Request to the collection.
 *
 * Parameters:
 *  pdisp    IDispatch of xrt data object to be added to collection.
 *
 */

STDMETHODIMP_(LPDISPATCH)   
CRequests::Add(VARIANT Name, VARIANT vdispProps)
{ 
    CRequest FAR* pRequest = NULL;  
    LPDISPATCH pdispRequest = NULL;
    long lPos = 0L;
   	CProperty FAR *pProperty = NULL;
	CProperties FAR *pSoloProperties = NULL;
	LPDISPATCH pdispProperties = NULL;
	VARIANT vXRTProps;
	VARIANT vXRTPropsVar;

	TRY{
		// Check Name parameter
		if(Name.vt == VT_EMPTY || Name.vt == VT_ERROR 
				|| Name.bstrVal == NULL){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_INVALIDREQUESTNAME
					, NULL); 
			m_bRaiseException = TRUE;
			return NULL;
		}

		// If a Maximum Requests is set
		// Make sure collection is not full
		// Else just keep adding
    	if (m_pDataObject->m_dwMaxRequests 
    			&& m_col.m_cElements + 1 > m_pDataObject->m_dwMaxRequests){     
			
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_COLLECTIONISFULL
					, NULL); 
					
			m_bRaiseException = TRUE;
			
			return NULL;
    	}

		if(FAILED(CheckPointer((LPVOID) Name.bstrVal, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 

			m_bRaiseException = TRUE;
			
			return NULL;
		}
		

		if(FAILED(CheckVariant(&vdispProps, TRUE))){
    		AddInDebugOutput(NULL, TEXT("CRequests::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 

			m_bRaiseException = TRUE;
			
			return NULL;
		}
		
		//Create a new Request
    	if (FAILED(CRequest::Create(Name.bstrVal, &pRequest))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_REQUESTCREATEFAIL
					, NULL); 
			MYRAISEEXCEPTION;
    	}       

    	//Get Request IDispatch
    	pRequest->QueryInterface(IID_IDispatch, (void FAR* FAR*) &pdispRequest); 

		// Save the index of the request    
    	pRequest->m_dwIndex = m_col.m_cElements;  

		// Add request to our requests collection
		if(FAILED(m_col.AddObject(pdispRequest, Name.bstrVal, FALSE, 0))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_ADDOBJECTTOCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}
		
		// Save parent DataObject
		pRequest->m_pDataObject = m_pDataObject;

		// Create a properties sub-object
    	if (FAILED(CProperties::Create(&pRequest->m_pProperties))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_PROPSCREATEFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}                                                                     
	   	
	   	pRequest->m_pProperties->m_pDataObject = m_pDataObject;

	   	pRequest->m_pProperties->m_pRequest = pRequest;

	   	pRequest->m_pRequests = this;

		// Add a Row to the Database for this new request
		if(FAILED(m_pDataObject->m_pCache->AddRequest(Name.bstrVal
				, pRequest, &pRequest->m_pxlRequest))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_ADDOBJECTTOCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

		// If we were sent some proeprties to add to the request,
		// copy the props over to this properties buffer which should
		// also assign them an XLOPER in cache.
		if(vdispProps.vt & VT_DISPATCH){    
			// Get IProperties from Properties parameter 	
			
			if(vdispProps.vt & VT_BYREF){
				pdispProperties = *(vdispProps.ppdispVal);
			}else{
				pdispProperties = vdispProps.pdispVal;
			}                                      

			if( FAILED( pdispProperties->QueryInterface(IID_IProperties
					, (LPVOID FAR *) &pSoloProperties) )){
				AddInDebugOutput(&m_excepinfo
						, TEXT("CRequests::Add")
						, IDS_DEBUG_INVALIDPROPERTIESPARAM
						, NULL); 
				MYRAISEEXCEPTION;
			}
	    
		    // Replicate properties
			if(FAILED(pRequest->m_pProperties->AddExistingProps(pSoloProperties))){
				AddInDebugOutput(&m_excepinfo
						, TEXT("CRequests::Add")
						, IDS_DEBUG_ADDEXISITINGPROPSFAILED
						, NULL); 
				MYRAISEEXCEPTION;
			}
			
			// Prepare Properties for XRTAdd
			VariantInit(&vXRTProps);
			VariantInit(&vXRTPropsVar);
			vXRTProps.vt = VT_DISPATCH;
			vXRTProps.pdispVal = pSoloProperties->m_pXRTDispatch;
			vXRTPropsVar.vt = VT_VARIANT | VT_BYREF;
			vXRTPropsVar.pvarVal = &vXRTProps;
			
			//Release from earlier QI			
			pSoloProperties->Release();
			pSoloProperties = NULL;
        }
        
		// Pass-on XRT Add		
		if(FAILED(XRTAdd(&Name, &vXRTPropsVar, &pRequest->m_pXRTDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_CALLTOXRTREQUESTSADDFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

    	// Get the properties dispatch from XRT object. It should have 
    	// created one when we called XRTAdd or when we call this...
	    if (FAILED(pRequest->XRTProperties(&pRequest->m_pProperties->m_pXRTDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_CALLTOXRTREQUESTPROPERTIESFAILED
					, NULL); 
			MYRAISEEXCEPTION;
	   	} 
    
    	//Save parent DataObject
		pRequest->m_pDataObject = m_pDataObject;
		
		if(vdispProps.vt & VT_DISPATCH){    
			VARIANT v;
			CProperty FAR *pProperty = NULL;
			ULONG i;
			CNode FAR *pNode = NULL;    
			
			VariantInit(&v);			    
			    
			// Assign my objects of type property to XRT's
			for(i=0, pNode = pRequest->m_pProperties->m_col.m_pFirstNode
					; i < pRequest->m_pProperties->m_col.m_cElements 
					; pNode = pNode->m_pNext, i++){
				
				pNode->m_pDispatch->QueryInterface(IID_IProperty
						, (LPVOID FAR *) &pProperty);	
			    
			    v.vt = VT_BSTR;
			    v.bstrVal = SysAllocString(pProperty->m_bstrName);
			    
			    pRequest->m_pProperties->XRTItem(&v, &pProperty->m_pXRTDispatch);
			    
			    VariantClear(&v);

				pProperty->Release();
				pProperty = NULL;
			}
		}	
		            
		pdispProperties->Release();		            
		pdispProperties = NULL;		            
		            
		// Release orginal request    
	    pRequest->Release();
	    pRequest = NULL;

    	return pdispRequest;
	}

	
	EXCEPT{
		if(pRequest)
			delete pRequest;

       	m_bRaiseException = TRUE;

    	if(pdispRequest)
			pdispRequest->Release();

		if(pSoloProperties)
			pSoloProperties->Release();

		if(pdispProperties)
			pdispProperties->Release();

   		AddInDebugOutput(&pRequest->m_excepinfo, TEXT("CRequests::Add")
   			, IDS_DEBUG_EXCEPTION, NULL); 

   		return NULL;   
	}
}



/*
 * CRequests::Add
 *
 * Purpose:
 *  Adds an Request to the collection.This Add is for used when cache defines collection.
 *  Usually because DataObjects.FindObjectByIndex was used.
 *
 * Parameters:
 *  pdisp    IDispatch of xrt data object to be added to collection.
 *
 */
HRESULT CRequests::Add(LPXLOPER pRow, BSTR bstrRequestName, DWORD dwID
		, CRequest FAR * FAR *ppRequest, CNode FAR* FAR* ppNode)
{ 
    CRequest FAR* pRequest = NULL;  
    long lPos = 0L;
   	CProperty FAR *pProperty = NULL;
	CProperties FAR *pSoloProperties = NULL;
	VARIANT vParam;
	LPDISPATCH pDispatch = NULL;
	CNode FAR* pNode = NULL;
	
	TRY{
		if(FAILED(CheckPointer((LPVOID) pRow, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckPointer((LPVOID) bstrRequestName, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckPointer((LPVOID) ppRequest, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppRequest = NULL;
	
		if(pNode != NULL){
			if(FAILED(CheckPointer((LPVOID) ppNode, 0))){
    			AddInDebugOutput(NULL, TEXT("CRequests::Add")
    				, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
		
			*ppNode = NULL;
		}
	
		// Make sure collection is not full    
    	if (m_pDataObject->m_dwMaxRequests 
    			&& m_col.m_cElements + 1 > m_pDataObject->m_dwMaxRequests){     
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_COLLECTIONISFULL
					, NULL); 
			MYRAISEEXCEPTION;
    	}
                   
		//Create a new Request
    	if (CRequest::Create(bstrRequestName, &pRequest)){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_REQUESTCREATEFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}       

		pRequest->QueryInterface(IID_IRequest, (LPVOID FAR *) &pDispatch);

		// Save the index of the request
    	pRequest->m_dwIndex = m_col.m_cElements;  
    
    	// Add new Request to safe array
		if(FAILED(m_col.AddObject(pDispatch, bstrRequestName, TRUE
				, dwID, &pNode))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_ADDOBJECTTOCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}                                  
		
		pDispatch->Release(); // Release from QI
		pDispatch = NULL;
		
		if(ppNode) 
			*ppNode = pNode;
		
    	// Get XRT Request
		VariantInit(&vParam);
		vParam.vt = VT_BSTR;
		vParam.bstrVal = SysAllocString(bstrRequestName);
	
    	if (FAILED(XRTItem(&vParam, &pRequest->m_pXRTDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_CALLTOXRTREQUESTSITEMFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}       
    
    	VariantClear(&vParam);
    
		// Save parent DataObject
		pRequest->m_pDataObject = m_pDataObject;

	   	pRequest->m_pRequests = this;

	    if (CProperties::Create(&pRequest->m_pProperties)){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_PROPERTIESCREATEFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}                                                                     

	   	pRequest->m_pProperties->m_pRequest = pRequest;

		pRequest->m_pProperties->m_pDataObject = pRequest->m_pDataObject;		

		pRequest->m_pxlRequest = pRow;	
		// Save row's respective request.				     
	
    	// Get the properties dispatch from XRT object. It should have 
	    // created one when we called XRTAdd.
	    if (pRequest->XRTProperties(&pRequest->m_pProperties->m_pXRTDispatch)){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Add")
					, IDS_DEBUG_CALLTOXRTREQUESTPROPERTIESFAILED
					, NULL); 
			MYRAISEEXCEPTION;
   		} 
    
		//Add Mirror DataItem
		//m_pMirrorDataItems->Add(vParam, pRequest->m_pProperties);
    
		VariantClear(&vParam);                      
                      
		*ppRequest = pRequest;
	                      
    	// Earlier we did a QI for this so no AddRef
    	return S_OK;
	}

	EXCEPT{
    	// Signal CRequests::Invoke to raise exception.
	    m_bRaiseException = TRUE;    
        
        if(pDispatch)
        	pDispatch->Release();
        
		//Clean Up
		if(pRequest)
			delete pRequest;
	    
		if(pSoloProperties)
			pSoloProperties->Release;	    
	    
		VariantClear(&vParam);                      

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::Add")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	
    	return ResultFromScode(E_FAIL);   
	}
}



/*
 * CRequests::Remove
 *
 * Purpose:
 *  Removes specified item from collection. Signals CRequests::Invoke to raise
 *  exception on ResultFromScode(E_FAIL).
 *
 * Parameters:   
 *   vIndex   Index of item to be removed. This index can be an integer or a string.
 *
 */
STDMETHODIMP_(void)
CRequests::Remove(VARIANT vParam)
{
	ULONG lIndex;
	LPDISPATCH pDispatch = NULL;
	HRESULT hr;

	TRY{
		if(FAILED(CheckVariant(&vParam, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Remove")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
			m_bRaiseException = TRUE;
			return;
		}
	
		if(FAILED(XRTRemove(&vParam))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Remove")
					, IDS_DEBUG_CALLTOXRTREQUESTSREMOVE
					, NULL); 
  			MYRAISEEXCEPTION;
		}
		
		// Find object based on VARIANT index
   		if(vParam.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) vParam.iVal, &pDispatch);
			lIndex = (long) vParam.iVal;
		}else if(vParam.vt == VT_I4){
			hr = m_col.FindObjectByIndex(vParam.lVal, &pDispatch);
			lIndex = (long) vParam.lVal;
   		}else if(vParam.vt == VT_BSTR){
			hr = m_col.FindObject(vParam.bstrVal, &pDispatch, &lIndex);
		}else{
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Remove")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
  			MYRAISEEXCEPTION;
		}

		// Check index                                  
		if(FAILED(hr)){        
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Remove")
					, IDS_DEBUG_FINDOBJECTINCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}
		
		//////////////////////////////////////////////////////////
		// Update the cache
		// Should probably move this code to CCache::RemoveRequest 
		if(m_col.m_cElements == 0){
        	return;
		}else if(m_col.m_cElements == 1){                 
			CRequest FAR *pRequest = NULL;
		
			pDispatch->QueryInterface(IID_IRequest, (LPVOID FAR *) &pRequest);

			for(DWORD i=0;i<m_pDataObject->m_dwMaxProperties;i++){
				pRequest->m_pxlRequest[i].xltype = xltypeNil;			
			}

			pRequest->Release();
			pRequest = NULL;
		}else{
			CRequest FAR *pRequest = NULL;
		
			pDispatch->QueryInterface(IID_IRequest, (LPVOID FAR *) &pRequest);
		
			LPVOID pSrc = (LPVOID) (&pRequest->m_pxlRequest[m_pDataObject->m_dwMaxProperties]);
			LPVOID pDest = (LPVOID) pRequest->m_pxlRequest;
			DWORD dwSize = ((m_col.m_cElements * m_pDataObject->m_dwMaxProperties * sizeof(XLOPER)) - (LONG) pSrc);
		
			_fmemcpy(pDest, pSrc, (size_t) dwSize);

			DWORD dwLimit = (m_col.m_cElements-1) * m_pDataObject->m_dwMaxProperties;
			
			for(DWORD i=m_col.m_cElements * m_pDataObject->m_dwMaxProperties;i>=dwLimit;i--){
				m_pDataObject->m_pCache->m_pxlCache[i].xltype = xltypeNil;			
			}

			pRequest->Release();
			pRequest = NULL;
		}
		//////////////////////////////////////////////////////////
				
		// Update collection
		if(FAILED(m_col.RemoveObject(pDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Remove")
					, IDS_DEBUG_REMOVEOBJECTFROMCOLLECTIONFAILED
					, NULL); 
  			MYRAISEEXCEPTION;
		}
		
		//Release from FindObject
		pDispatch->Release();
		pDispatch = NULL;
	
	    return; 
	}

	EXCEPT{
		if(pDispatch)
			pDispatch->Release();
						
		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::Remove")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	
		return;
	}
}

/*
 * CRequests::Object - Read-only property
 *
 * Purpose:
 *	Returns WOSA/XRT object CProperty object. Used by VBA developer to call
 *	properties and methods which are specific to the WOSA/XRT object.
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *)
CRequests::get_Object(){
	if(FAILED(CheckDispatch(m_pXRTDispatch))){
   		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
   			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 

		m_bRaiseException = TRUE;
		return NULL;
	}

	m_pXRTDispatch->AddRef();	
                   
	return m_pXRTDispatch;                     
}

/*
 * CRequests::Item
 *
 * Purpose:
 *	Return specific Request from collection
 *
 * Return Value:
 *  LPDISATCH
 *
 */

STDMETHODIMP_(LPDISPATCH)
CRequests::get_Item(VARIANT Index)  
{   
	LPDISPATCH pDispatch;
	HRESULT hr;

	TRY{
		if(FAILED(CheckVariant(&Index, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::get_Item")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
			m_bRaiseException = TRUE;
			return NULL;
		}
	
		// check index    
   		if(Index.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) Index.iVal, &pDispatch);
   		}else if(Index.vt == VT_I4){
	   		hr = m_col.FindObjectByIndex(Index.lVal, &pDispatch);
   		}else if(Index.vt == VT_BSTR){
	   		hr = m_col.FindObject(Index.bstrVal, &pDispatch, NULL);
   		}else{
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::g_Item")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
  			MYRAISEEXCEPTION;
	   	}

		VariantClear(&Index);
	
    	return pDispatch;
	}
	
	EXCEPT{
		m_bRaiseException = TRUE;
          
   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::get_Item")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	
		return NULL;
	}
}

/*
 * CRequests::DataObject
 *
 * Purpose:
 *	Returns the parent DataObject
 *
 * Return Value:
 *  LPDISATCH
 *
 */


STDMETHODIMP_(IDispatch *)
CRequests::get_DataObject(){                                                                            
	LPDISPATCH pDispatch;
	
	m_pDataObject->QueryInterface(IID_IDispatch, (LPVOID FAR *)&pDispatch);
	                             
	//QI does AddRef                             
	                             
	return pDispatch;
}


/*
 * CRequests::get_NewEnum
 *
 * Purpose:
 *  Returns an enumerator (IEnumVARIANT) for the items curently in the collection. 
 *  The NewEnum property is restricted and so is invisible to users of an 
 *  automation controller's scripting language. Automation controllers that support
 *  a 'For Each' statement to iterate through the elements of a collection will use
 *  the enumerator returned by NewEnum. The enumerator creates a snapshot of the
 *  the current state of the collection.
 *
 */
STDMETHODIMP_(LPUNKNOWN)
CRequests::get__NewEnum(void)
{
    LPUNKNOWN punkEnumVariant = NULL;

	TRY{
		if(FAILED(m_col.GetEnum(&punkEnumVariant))){
			MYRAISEEXCEPTION;	
		}
		
    	return punkEnumVariant; 
	}
    
	EXCEPT{
    	m_bRaiseException = TRUE;    
		
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::_NewEnum")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;    	
	}
}

/*
 * CRequests::Count
 *
 * Purpose:
 *	Returns number of request objects in collection
 *
 * Return Value:
 *  LPDISATCH
 *
 */

STDMETHODIMP_(long) CRequests::get_Count (void){
	return m_col.m_cElements;
}

/*
 * CRequests::Reset
 *
 * Purpose:
 *	Resets collection to zero
 *
 * Return Value:
 *  void
 *
 */

STDMETHODIMP_(void) CRequests::Reset(void){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Reset");
	LONG lIndex;
	CRequest FAR *pRequest = NULL;
	VARIANT vResult, v;
	
	TRY{
		VariantInit(&vResult);
		VariantInit(&v);
	
		HRESULT hr = InvokeMethodHelper(m_pXRTDispatch, &dispid, 
			&dispparams, szMember, NULL);
                
		if(FAILED(hr)){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::Reset")
					, IDS_DEBUG_CALLTOXRTREQUESTSRESETFAILED
					, NULL); 
			MYRAISEEXCEPTION;
			return;                
		}
                
		for (lIndex = 0; lIndex < (long)m_col.m_cElements; lIndex++){
			m_col.RemoveObject(pRequest);
		}

		return;
	}

	EXCEPT{
		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::Reset")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return;		
	}
}

/*
 * CRequests::CreateProperties
 *
 * Purpose:
 *	Creates and initializes Properties collection
 *
 * Return Value:
 *  LPDISPATCH
 *
 */

STDMETHODIMP_(IDispatch *)
CRequests::CreateProperties (VARIANT property1, VARIANT property2
		, VARIANT property3, VARIANT property4, VARIANT property5
		, VARIANT property6, VARIANT property7, VARIANT property8)
{
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("CreateProperties");
	VARIANT params[8];
	DISPPARAMS dispparams;

	CProperties FAR *pProperties = NULL;
	LPDISPATCH pDispatch = NULL;
	CProperty FAR *pProperty = NULL;
	VARIANT vResult;

	int nParamCount = 0;
	int i = 0;
	
	TRY{    
		// Init params
		for(i=0;i<8;i++){	
			VariantInit(&params[i]);
		}

		VariantInit(&vResult);
	
		// *************************** Create paramters	****************************
		if(property1.vt != VT_ERROR && property1.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[7], &property1))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property2.vt != VT_ERROR && property2.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[6], &property2))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property3.vt != VT_ERROR && property3.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[5], &property3))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property4.vt != VT_ERROR && property4.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[4], &property4))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property5.vt != VT_ERROR && property5.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[3], &property5))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property6.vt != VT_ERROR && property6.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[2], &property6))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property7.vt != VT_ERROR && property7.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[1], &property7))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		if(property8.vt != VT_ERROR && property8.vt != VT_EMPTY){
			if(FAILED(VariantCopy(&params[0], &property8))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CANNOTCOPYPROPERTY
					, NULL); 
				MYRAISEEXCEPTION;
			}
			nParamCount++;		
		}
	
		dispparams.rgvarg = params;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = nParamCount;
		dispparams.cNamedArgs = 0;     
	
// *************************************************************************

		// Call XRT object's CreateProperties
		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){
		
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CALLTOXRTREQUESTSCREATEPROPSFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

		if(FAILED(CheckVariant(&vResult, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_BADVARIANT, NULL); 
			MYRAISEEXCEPTION;
		}

		// Create our collection for proeprties	
		if(FAILED(CProperties::Create(&pProperties))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CRequests::CreateProperties")
					, IDS_DEBUG_CREATEPROPSFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

		// Save dispatch	
		pProperties->m_pXRTDispatch = vResult.pdispVal;                     

		// AddRef properties

		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		pProperties->m_pXRTDispatch->AddRef();
	
		// Set parent
		pProperties->m_pDataObject = m_pDataObject;

		// Get our properties dispatch
		// Internal QI; just a recast
    	pProperties->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pDispatch);

		// Add Name as first property in properties	                   
		VariantClear(&vResult);
		vResult.vt = VT_BSTR;
		vResult.bstrVal = SysAllocString(TEXT("Name"));

		pProperties->Add(vResult, &pProperty, 0L);

		// Add will return addref'd property because it think its for VB
		// its already been addref'd once for the collection so we release 
		// here. We definately want this here!!!!!
		pProperty->Release();
		pProperty = NULL;
		

		VariantClear(&vResult);
	
		// Add the rest of the proeprties backwards
		for(i=7;i>=0;i--){	
			if(params[i].vt != VT_EMPTY && params[i].vt != VT_ERROR){ 
				pProperties->Add(params[i], NULL, 0L);
			}
		}

		// Clean up variants	
		for(i=0;i<8;i++){	
			VariantClear(&params[i]);
		}

		pProperties->Release();
		pProperties = NULL;
	
		// Return our proeprties' dispatch	
		return pDispatch;
	}
	
	EXCEPT{
		for(i=0;i<8;i++){	
			VariantClear(&params[i]);
		}

		if(pProperty)
			pProperty->Release();			

		if(pProperties == NULL)
			return NULL;
			
		pProperties->Reset();
	
		if(pProperties)
			delete pProperties;

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::CreateProperties")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		m_bRaiseException = TRUE;
		
		return NULL;
	}
}


/*
 * CRequests::XRTItem
 *
 * Purpose:
 *	Calls CRequests.Item on real WOSA/XRT object
 *
 * Return Value:
 *  LPDISATCH
 *
 */

HRESULT CRequests::XRTItem(VARIANT FAR *pvIndex, LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Item");
	VARIANTARG params[1];
	DISPPARAMS dispparams;
	VARIANT vResult;

	TRY{
		if(FAILED(CheckVariant(pvIndex, TRUE))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::XRTItem")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		*ppDispatch = NULL;
			
		VariantInit(&vResult);			
		VariantInit(&params[0]);			
		
		if(FAILED(VariantCopy(&params[0], pvIndex))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTItem")
    			, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
			return ResultFromScode(E_FAIL);
		}

		dispparams.rgvarg = params;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;

		if(FAILED(InvokePropGetHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTItem")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}		

		if(FAILED(CheckVariant(&vResult, TRUE))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTItem")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}

		*ppDispatch	= vResult.pdispVal;

		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();

		return S_OK;
	}
	
	EXCEPT{
		VariantClear(&vResult);			
		VariantClear(&params[0]);			

		return ResultFromScode(E_FAIL);	
	}
}

/*
 * CRequests::XRTAdd
 *
 * Purpose:
 *	Calls CRequests.Add on real WOSA/XRT object
 *
 * Return Value:
 *  LPDISATCH
 *
 */

HRESULT CRequests::XRTAdd(VARIANT FAR *pvName, VARIANT FAR *pvProperties
		, LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Add");
	VARIANT params[2];
	DISPPARAMS dispparams;
	LPTSTR pszMember = szMember;
	VARIANT vResult;

	TRY{
		if(FAILED(CheckVariant(pvName, TRUE))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckVariant(pvProperties, FALSE))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppDispatch = NULL;
	
		VariantInit(&vResult);			
		VariantInit(&params[0]);			
		VariantInit(&params[1]);			

		if(FAILED(VariantCopy((VARIANT *) &params[1], pvName))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
			return ResultFromScode(E_FAIL);
		}
					  
		if(FAILED(VariantCopy((VARIANT *) &params[0], pvProperties))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
			VariantClear(&params[1]);
			return ResultFromScode(E_FAIL);
		}
		
		dispparams.rgvarg = (VARIANT *) params;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 2;
		dispparams.cNamedArgs = 0;

		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){

	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   				, IDS_DEBUG_EXCEPTION, NULL); 

			MYRAISEEXCEPTION;
		}

		if(FAILED(CheckVariant(&vResult, TRUE))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}

		*ppDispatch = vResult.pdispVal;
		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();

		VariantClear((VARIANT *)&params[0]);			
		VariantClear((VARIANT *)&params[1]);			

		return S_OK;
	}

	EXCEPT{
		VariantClear((VARIANT *)&params[0]);			
		VariantClear((VARIANT *)&params[1]);			
		VariantClear((VARIANT *)&vResult);			

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}


/*
 * CRequests::XRTRemove
 *

 * Purpose:
 *	Calls CRequests.Remove on real WOSA/XRT object
 *
 * Return Value:
 *  LPDISATCH
 *
 */

HRESULT CRequests::XRTRemove(VARIANT FAR *pvIndex){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Remove");
	VARIANT param;
	DISPPARAMS dispparams;
	VARIANT vResult;

	TRY{    
		if(FAILED(CheckVariant(pvIndex, TRUE))){
			return ResultFromScode(E_FAIL);
		}
	
		VariantInit(&param);
		VariantInit(&vResult);

		if(FAILED(VariantCopy(&param, pvIndex))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_INVALIDPARAMETERS, NULL); 

			return ResultFromScode(E_FAIL);
		}
		
		dispparams.rgvarg = &param;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;
	
		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){

			VariantClear(&param);

	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTRemove")
   				, IDS_DEBUG_EXCEPTION, NULL); 

			return ResultFromScode(E_FAIL);
		}
		
		VariantClear(&param);
	
		return S_OK;
	}

	#ifdef _WIN32
	EXCEPT{
		VariantClear(&param);
		VariantClear(&vResult);

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTRemove")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);;
	}
	#endif
}
