// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		This module is orginally from MFC. It has been altered to do 
//		auto-registration for the add-in.
//
// 		This is a part of the Microsoft Foundation Classes C++ library.
// 		Copyright (C) 1992 Microsoft Corporation
// 		All rights reserved.
//
// 		This source code is only intended as a supplement to the
// 		Microsoft Foundation Classes Reference and Microsoft
// 		QuickHelp and/or WinHelp documentation provided with the library.
// 		See these sources for detailed information regarding the
// 		Microsoft Foundation Classes product.

#include "precomp.h"
#include <shellapi.h>
#include "xrtxlreg.h"
#include "guid.h"

//////////////////////////////////////////////////////////////////////////////
// data for UpdateRegistry functionality

static const TCHAR sz00[] = TEXT("%6\0") TEXT("%2");
static const TCHAR sz01[] = TEXT("%6\\CLSID\0") TEXT("%1");
static const TCHAR sz02[] = TEXT("%4\0") TEXT("%2");
static const TCHAR sz03[] = TEXT("%4\\CLSID\0") TEXT("%1");
static const TCHAR sz04[] = TEXT("CLSID\\%1\0") TEXT("%2");
static const TCHAR sz05[] = TEXT("CLSID\\%1\\ProgID\0") TEXT("%4");
static const TCHAR sz06[] = TEXT("CLSID\\%1\\TypeLib\0") TEXT("%5");
#ifndef _WIN32
static const TCHAR sz07[] = TEXT("CLSID\\%1\\InProcServer\0") TEXT("%3");
#else
static const TCHAR sz07[] = TEXT("CLSID\\%1\\InProcServer32\0") TEXT("%3");
#endif
static const TCHAR sz08[] = TEXT("CLSID\\%1\\VersionIndependentProgID\0") TEXT("%6");
static const TCHAR sz09[] = TEXT("Typelib\\%5\0");
static const TCHAR sz10[] = TEXT("Typelib\\%5\\1.0\0") TEXT("%2");
#ifndef _WIN32
static const TCHAR sz11[] = TEXT("Typelib\\%5\\1.0\\409\\win16\0") TEXT("%3");
#else
static const TCHAR sz11[] = TEXT("Typelib\\%5\\1.0\\409\\win32\0") TEXT("%3");
#endif

// registration for OAT_SERVER
static const LPCTSTR rglpszServerRegister[] =
{
	sz00, sz01, sz02, sz03, sz04, sz05, sz06, sz07, sz08, sz09, sz10, sz11, NULL
};

static const LPCTSTR rglpszServerOverwrite[] =
{
	sz01, sz03, sz05, sz06, sz07, sz08, sz09, sz10, sz11, NULL
};

struct STANDARD_ENTRY
{
	const LPCTSTR* rglpszRegister;
	const LPCTSTR* rglpszOverwrite;
};

static const STANDARD_ENTRY rgStdEntry[] =
{
	{ rglpszServerRegister, rglpszServerOverwrite }
};

/////////////////////////////////////////////////////////////////////////////
// Special registration for apps that wish not to use REGLOAD

BOOL WINAPI OleRegisterServerClass(	
	REFCLSID clsid, LPTSTR lpszClassName
	, LPTSTR lpszProgID, LPTSTR lpszVIProgID
	, REFCLSID tylibclsid
	, LPTSTR* rglpszRegister, LPTSTR* rglpszOverwrite)
{
	// use standard registration entries if non given
	if (rglpszRegister == NULL)
		rglpszRegister = (LPTSTR*) rgStdEntry->rglpszRegister;
	if (rglpszOverwrite == NULL)
		rglpszOverwrite = (LPTSTR*)rgStdEntry->rglpszOverwrite;

	LPTSTR lpszClassID;
	LPTSTR lpszTypeLibClassID;
	LPTSTR rglpszSymbols[8];
	
	::StringFromCLSID(clsid, &lpszClassID);
	::StringFromCLSID(tylibclsid, &lpszTypeLibClassID);
	
	rglpszSymbols[0] = lpszClassID;
	rglpszSymbols[1] = lpszClassName;

	// get path name to server
	TCHAR szPathName[MAX_PATHNAME_LEN];

	#ifdef _WIN32
		HMODULE hMod = GetModuleHandle(FILENAME);
	#else
		HMODULE hMod = g_hinst ;
	#endif

	::GetModuleFileName(hMod, szPathName, 256);
	rglpszSymbols[2] = szPathName;

	// fill in rest of symbols
	rglpszSymbols[3] = lpszProgID;
	rglpszSymbols[4] = lpszTypeLibClassID;
	rglpszSymbols[5] = lpszVIProgID;

	// update the registry with helper function
	BOOL bResult;
	bResult = OleRegisterHelper(rglpszRegister, (LPTSTR *) rglpszSymbols
			, 6, FALSE);

	if (bResult && rglpszOverwrite != NULL)
		bResult = OleRegisterHelper(rglpszOverwrite, (LPTSTR *) rglpszSymbols
			, 6, TRUE);

	return bResult;
}

// writes key/value pairs to system registry
BOOL WINAPI OleRegisterHelper(LPTSTR* rglpszRegister,
		LPTSTR* rglpszSymbols, int nSymbols, BOOL bReplace)
{
	TCHAR strKey[MAX_REG_KEY_LEN];
	TCHAR strValue[MAX_REG_VALUE_LEN];
	
	// keeping a key open makes this go a bit faster
	HKEY hKeyTemp = NULL;
	RegOpenKey(HKEY_CLASSES_ROOT, TEXT("CLSID"), &hKeyTemp);

	BOOL bResult = TRUE;
	while (*rglpszRegister != NULL)
	{
		LPCTSTR lpszKey = *rglpszRegister++;
		if (*lpszKey == '\0')
			continue;

		LPCTSTR lpszValue = lpszKey + lstrlen(lpszKey) + 1;
	
		memset(strKey, 0, sizeof(strKey));
		memset(strValue, 0, sizeof(strValue));

		FormatStrings(strKey, lpszKey, (LPCTSTR *) rglpszSymbols, nSymbols);
		FormatStrings(strValue, lpszValue, (LPCTSTR *) rglpszSymbols, nSymbols);

		if (!bReplace)
		{
			TCHAR szBuffer[MAX_REG_VALUE_LEN];
			LONG lSize = sizeof(szBuffer);
			if (::RegQueryValue(HKEY_CLASSES_ROOT, strKey, szBuffer, &lSize) ==
				ERROR_SUCCESS)
			{
#ifdef _DEBUG
				if (!lstrcmp(strValue, szBuffer))
				{

				}
#endif
				continue;
			}
		}

		if (::RegSetValue(HKEY_CLASSES_ROOT, strKey, REG_SZ, strValue, 0)
			!= ERROR_SUCCESS)
		{
			bResult = FALSE;
			break;
		}
	}

	if (hKeyTemp != NULL)
		RegCloseKey(hKeyTemp);

	return bResult;
}

void WINAPI FormatStrings(LPCTSTR lpszString, LPCTSTR lpszFormat,
		LPCTSTR* rglpsz, int nString)
{
	// determine length of destination string
	int nTotalLen = 0;
	LPCTSTR pchSrc = lpszFormat;
	while (*pchSrc != '\0')
	{
		if (pchSrc[0] == '%' && (pchSrc[1] >= '1' && pchSrc[1] <= '9'))
		{
			int i = pchSrc[1] - '1';
			pchSrc += 2;
			if (i >= nString)
				++nTotalLen;
			else if (rglpsz[i] != NULL)
				nTotalLen += lstrlen(rglpsz[i]);
		}
		else
		{
			if (IsDBCSLeadByte((BYTE)*pchSrc))
				++nTotalLen, ++pchSrc;
			++pchSrc;
			++nTotalLen;
		}
	}

	pchSrc = lpszFormat;
	LPTSTR pchDest = (LPTSTR) lpszString;
	while (*pchSrc != '\0')
	{
		if (pchSrc[0] == '%' && (pchSrc[1] >= '1' && pchSrc[1] <= '9'))
		{
			int i = pchSrc[1] - '1';
			pchSrc += 2;
			if (i >= nString)
			{
				*pchDest++ = '?';
			}
			else if (rglpsz[i] != NULL)
			{
				lstrcpy(pchDest, rglpsz[i]);
				pchDest += lstrlen(pchDest);
			}
		}
		else
		{
			if (IsDBCSLeadByte((BYTE)*pchSrc))
				*pchDest++ = *pchSrc++; // copy first of 2 bytes
			*pchDest++ = *pchSrc++;
		}
	}
}
